# The ADMM denoiser with L1-norm prior

# import packages
import numpy as np
from skimage.metrics import peak_signal_noise_ratio

from pathlib import Path
from tqdm import tqdm
from itertools import product

def denoiser_admm_l1norm(b, gt, lam, rho, num_iters):
    """
    b: The noisy Poissonian measurement
    gt: The ground truth image
    lam: The relative weight for the L1-norm term
    rho: The rho parameter for ADMM
    num_iter: The number of iterations for ADMM
    The ground truth image is only used for calculating the PSNR after each iteration
    The function returns the denoised image and the PSNR values after each iteration
    """

    # Prelocate memory for the PSNR values after each iteration
    PSNR = np.zeros((num_iters,1))

    # initialize x,z,u with all zeros
    x = np.zeros_like(b)
    z1 = np.zeros_like(b)
    z2 = np.zeros_like(b)
    z3 = np.zeros_like(b)
    u1 = np.zeros_like(b)
    u2 = np.zeros_like(b)
    u3 = np.zeros_like(b)

    for it in tqdm(range(num_iters)):

        # x update - simply the average of 3 v's
        v1 = z1 - u1
        v2 = z2 - u2
        v3 = z3 - u3
        x = (v1 + v2 + v3)/3

        # z1 update - element-wise quadratic root
        v = x + u1
        z1 = - (1-rho*v)/2/rho + np.sqrt(((1-rho*v)/2/rho)**2 + b/rho)

        # z2 update - indicator function for non-negativity constrain
        v = x + u2
        z2 = np.maximum(0, v)

        # z3 update - soft thresholding
        kappa = lam / rho
        v = x + u3
        # proximal operator of anisotropic TV term
        z3 = np.maximum(1 - kappa/np.abs(v), 0) * v

        # u update
        u1 = u1 + x - z1
        u2 = u2 + x - z2
        u3 = u3 + x - z3

        # compute PSNR using skimage library
        PSNR[it] = round(peak_signal_noise_ratio(np.clip(x,0.0,1.0), gt), 3)

    return x, PSNR

