# The ADMM denoiser with TV prior

# import packages
import numpy as np
from numpy.fft import fft2, ifft2
from skimage.metrics import peak_signal_noise_ratio

from pypher.pypher import psf2otf
from pathlib import Path
from tqdm import tqdm
from itertools import product

def denoiser_admm_tv(b, gt, lam, rho, num_iters):
    """
    b: The noisy Poissonian measurement
    gt: The ground truth image
    lam: The relative weight for the TV term
    rho: The rho parameter for ADMM
    num_iter: The number of iterations for ADMM
    The ground truth image is only used for calculating the PSNR after each iteration
    The function returns the denoised image and the PSNR values after each iteration
    """

    # Prelocate memory for the PSNR values after each iteration
    PSNR = np.zeros((num_iters,1))

    # finite differences kernels and corresponding otfs
    dx = np.array([[-1., 1.]])
    dy = np.array([[-1.], [1.]])
    dxFT = psf2otf(dx, b.shape)
    dyFT = psf2otf(dy, b.shape)
    dxTFT = np.conj(dxFT)
    dyTFT = np.conj(dyFT)

    # initialize x,z,u with all zeros
    x = np.zeros_like(b)
    z1 = np.zeros_like(b)
    z2 = np.zeros_like(b)
    z3 = np.zeros_like(b)
    z4 = np.zeros_like(b)
    u1 = np.zeros_like(b)
    u2 = np.zeros_like(b)
    u3 = np.zeros_like(b)
    u4 = np.zeros_like(b)

    # precompute the denominator for the x-update
    denom = 1 + 1 + dxTFT*dxFT + dyTFT*dyFT

    for it in tqdm(range(num_iters)):

        # x update
        v1 = z1 - u1
        v2 = z2 - u2
        v3 = z3 - u3
        v4 = z4 - u4

        nom = fft2(v1) + fft2(v2) + dxTFT * fft2(v3) + dyTFT * fft2(v4) # The nominator
        x = np.real(ifft2(nom / denom))

        # z1 update - element-wise quadratic root
        v = x + u1
        z1 = - (1-rho*v)/2/rho + np.sqrt(((1-rho*v)/2/rho)**2 + b/rho)

        # z2 update - indicator function for non-negativity constrain
        v = x + u2
        z2 = np.maximum(0, v)

        # z3 update - soft thresholding
        kappa = lam / rho
        v = np.real(ifft2(fft2(x) * dxFT)) + u3
        # proximal operator of anisotropic TV term
        z3 = np.maximum(1 - kappa/np.abs(v), 0) * v

        # z4 update - soft thresholding
        kappa = lam / rho
        v = np.real(ifft2(fft2(x) * dyFT)) + u4
        # proximal operator of anisotropic TV term
        z4 = np.maximum(1 - kappa/np.abs(v), 0) * v

        # u update
        u1 = u1 + x - z1
        u2 = u2 + x - z2
        u3 = u3 + np.real(ifft2(fft2(x) * dxFT)) - z3
        u4 = u4 + np.real(ifft2(fft2(x) * dyFT)) - z4

        # compute PSNR using skimage library
        PSNR[it] = round(peak_signal_noise_ratio(np.clip(x,0.0,1.0), gt), 3)

    return x, PSNR

