# import packages
import numpy as np
from numpy.fft import fft2, ifft2

import skimage.io as io
from skimage.metrics import peak_signal_noise_ratio 
from skimage.color import rgb2gray

from pypher.pypher import psf2otf
from pathlib import Path
from tqdm import tqdm
from itertools import product
import matplotlib.pyplot as plt

# import the ADMM denoisers with L1-norm, TV or LLT priors
from denoiser_admm_tv import *
from denoiser_admm_l1norm import *
from denoiser_admm_llt import *

# Define the function to run the denoisers and plot the results
def Poisson_denoiser(b, gt, lam_array, rho_array, num_iters, filename):
    """
    b: The noisy Poissonian measurement
    gt: The ground truth image
    lam_array: 3*1 array. The relative weight for the L1-norm, TV and LLT term
    rho_array: 3*1 array. The rho parameter for ADMM with L1-norm, TV and LLT prior
    num_iter: The number of iterations for ADMM. Same for all three priors
    filename: The filename for saving the output figures
    """

    # compute the nosiy image PSNR using skimage library
    PSNR_noisy = round(peak_signal_noise_ratio(gt, b), 3)

    # ADMM+L1-norm denoiser parameters
    rho         = rho_array[0]    # rho parameter of admm
    lam         = lam_array[0]    # relative weight of L1-norm term

    # run ADMM+L1-norm denoiser
    x_admm_l1norm, PSNR_admm_l1norm = denoiser_admm_l1norm(b,gt,lam,rho,num_iters)
    # clip results to make sure it's within the range [0,1]
    x_admm_l1norm = np.clip(x_admm_l1norm,0.0,1.0)

    # ADMM+TV denoiser parameters
    rho         = rho_array[1]    # rho parameter of admm
    lam         = lam_array[1]    # relative weight of TV term

    # run ADMM+TV denoiser
    x_admm_tv, PSNR_admm_TV = denoiser_admm_tv(b,gt,lam,rho,num_iters)
    # clip results to make sure it's within the range [0,1]
    x_admm_tv = np.clip(x_admm_tv,0.0,1.0)

    # ADMM+LLT denoiser parameters
    rho         = rho_array[2]    # rho parameter of admm
    lam         = lam_array[2]    # relative weight of TV term

    # run ADMM+LLT denoiser
    x_admm_llt, PSNR_admm_llt = denoiser_admm_llt(b,gt,lam,rho,num_iters)
    # clip results to make sure it's within the range [0,1]
    x_admm_llt = np.clip(x_admm_llt,0.0,1.0)

    # show results
    # fig1 shows the original, noisy and denoised images
    fig1 = plt.figure(figsize=(75, 50))

    ax = fig1.add_subplot(2, 3, 1)
    ax.imshow(gt, cmap='gray')
    ax.set_title("Target Image", fontsize=60)
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)

    ax = fig1.add_subplot(2, 3, 2)
    ax.imshow(b, cmap='gray')
    ax.set_title(f'Noisy image, PSNR: {PSNR_noisy} dB', fontsize=60)
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)

    ax = fig1.add_subplot(2, 3, 4)
    ax.imshow(x_admm_l1norm, cmap='gray')
    ax.set_title(f'ADMM+L1-norm, PSNR: {PSNR_admm_l1norm[-1,0]} dB', fontsize=60)
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)

    ax = fig1.add_subplot(2, 3, 5)
    ax.imshow(x_admm_tv, cmap='gray')
    ax.set_title(f'ADMM+TV, PSNR: {PSNR_admm_TV[-1,0]} dB', fontsize=60)
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)

    ax = fig1.add_subplot(2, 3, 6)
    ax.imshow(x_admm_llt, cmap='gray')
    ax.set_title(f'ADMM+LLT, PSNR: {PSNR_admm_llt[-1,0]} dB', fontsize=60)
    ax.xaxis.set_visible(False)
    ax.yaxis.set_visible(False)

    # Save the images as svg files for easy rearrangement using Adobe Illustrator
    plt.savefig(filename+'_images.svg', bbox_inches = 'tight')

    # fig2 shows the PSNR values as a function of iteration for each method
    fig2, ax = plt.subplots(figsize=(5, 4))
    ax.plot(range(num_iters), PSNR_admm_l1norm, label='L1 norm')
    ax.plot(range(num_iters), PSNR_admm_TV, label='TV norm')
    ax.plot(range(num_iters), PSNR_admm_llt, label='LLT norm')
    ax.set_xlabel('Number of iteration', fontsize=12)
    ax.set_ylabel('PSNR (dB)', fontsize=12)
    ax.legend()

    plt.savefig(filename + '_iterations.svg', bbox_inches='tight')


#################################################################################
# The following part just loads different synthetic or experimental data and perform the denoising task
# Note: The rho and lam parameters have been manually optizmized near the local parameter space

# The interference pattern image (synthetic data)
name = 'interference'
filename = name + '_synthetic'
gt = io.imread(f'{name}.png').astype(float) / 255
gt = rgb2gray(gt)

# synthetic data with poisson noise
PEAK = 5
b = np.random.poisson(gt * PEAK) / PEAK

# ADMM+prior solver parameters
rho_array = np.array([[60], [10], [10]])   # rho parameter of admm with L1-norm, TV and LLT prior
lam_array = np.array([[0.004], [1], [1]])  # relative weight of L1-norm, TV and LLT term
num_iters = 75

# Perform the denoising task using ADMM + 3 different priors
Poisson_denoiser(b, gt, lam_array, rho_array, num_iters, filename)
#################################################################################

# The bird image (synthetic data)
name = 'birds'
filename = name + '_synthetic'
gt = io.imread(f'{name}.png').astype(float)/255
gt = rgb2gray(gt)

# synthetic data with poisson noise
PEAK = 5
b = np.random.poisson(gt * PEAK) / PEAK

# ADMM+prior solver parameters
rho_array = np.array([[50.0], [10], [10]])     # rho parameter of admm with L1-norm, TV and LLT prior
lam_array = np.array([[0.002], [0.5], [0.5]])  # relative weight of L1-norm, TV and LLT term
num_iters = 75

# Perform the denoising task using ADMM + 3 different priors
Poisson_denoiser(b, gt, lam_array, rho_array, num_iters, filename)
#################################################################################

# The USAF image (synthetic data)
name = 'USAF'
filename = name + '_synthetic'
gt = io.imread(f'{name}.jpg').astype(float)/255

# synthetic data with poisson noise
PEAK = 5
b = np.random.poisson(gt * PEAK) / PEAK

# ADMM+prior solver parameters
rho_array = np.array([[0.75], [5], [5]])       # rho parameter of admm with L1-norm, TV and LLT prior
lam_array = np.array([[0.005], [0.5], [0.5]])  # relative weight of L1-norm, TV and LLT term
num_iters = 75

# Perform the denoising task using ADMM + 3 different priors
Poisson_denoiser(b, gt, lam_array, rho_array, num_iters, filename)
#################################################################################

# The USAF image (experimental data)
b = io.imread('USAF_10ms_noisy.tif').astype(float)  # The noisy image with 10-ms exposure time
gt = io.imread('USAF_10s_gt.tif').astype(float) # The ground-truth image with 10-s exposure time
filename = 'USAF_experiment'
b = b * np.mean(gt) / np.mean(b)  # Normalize the mean intensity of two images.
b = b / 65535 * 3 # Normalize the image and scale it up so that the maximum is close to 1
gt = gt / 65535 * 3 # Normalize the image and scale it up so that the maximum is close to 1

# ADMM+prior solver parameters
rho_array = np.array([[40], [0.1], [0.1]])      # rho parameter of admm with L1-norm, TV and LLT prior
lam_array = np.array([[0.0001], [0.3], [0.3]])  # relative weight of L1-norm, TV and LLT term
num_iters = 75

# Perform the denoising task using ADMM + 3 different priors
Poisson_denoiser(b, gt, lam_array, rho_array, num_iters, filename)
#################################################################################

# The particle dark-field image (experimental data)
b = io.imread('particle_10ms_noisy.tif').astype(float)  # The noisy image with 10-ms exposure time
gt = io.imread('particle_10s_gt.tif').astype(float) # The ground-truth image with 10-s exposure time
filename = 'particle_experiment'
b = b * np.mean(gt) / np.mean(b)  # Normalize the mean intensity of two images.
b = b / 65535 * 20 # Normalize the image and scale it up so that the maximum is close to 1
gt = gt / 65535 * 20 # Normalize the image and scale it up so that the maximum is close to 1

# ADMM+prior solver parameters
rho_array = np.array([[5.], [40.], [40.]])  # rho parameter of admm with L1-norm, TV and LLT prior
lam_array = np.array([[0.005], [0.5], [0.5]])  # relative weight of L1-norm, TV and LLT term
num_iters = 75

# Perform the denoising task using ADMM + 3 different priors
Poisson_denoiser(b, gt, lam_array, rho_array, num_iters, filename)
#################################################################################

# The boat image (experimental data)
b = io.imread('boat_10ms_noisy.tif').astype(float)  # The noisy image with 10-ms exposure time
gt = io.imread('boat_10s_gt.tif').astype(float) # The ground-truth image with 10-s exposure time
filename = 'boat_experiment'
b = b * np.mean(gt) / np.mean(b)  # Normalize the mean intensity of two images.
b = b / 65535 * 3 # Normalize the image and scale it up so that the maximum is close to 1
gt = gt / 65535 * 3 # Normalize the image and scale it up so that the maximum is close to 1


# ADMM+prior solver parameters
rho_array = np.array([[0.05], [60.], [60.]])  # rho parameter of admm with L1-norm, TV and LLT prior
lam_array = np.array([[2.5], [2], [2]])       # relative weight of L1-norm, TV and LLT term
num_iters = 75

# Perform the denoising task using ADMM + 3 different priors
Poisson_denoiser(b, gt, lam_array, rho_array, num_iters, filename)
#################################################################################

# The level and tape image (experimental data)
b = io.imread('level_10ms_noisy.tif').astype(float)  # The noisy image with 10-ms exposure time
gt = io.imread('level_10s_gt.tif').astype(float) # The ground-truth image with 10-s exposure time
filename = 'level_experiment'
b = b * np.mean(gt) / np.mean(b)  # Normalize the mean intensity of two images.
b = b / 65535 * 4 # Normalize the image and scale it up so that the maximum is close to 1
gt = gt / 65535 * 4 # Normalize the image and scale it up so that the maximum is close to 1

# ADMM+prior solver parameters
rho_array = np.array([[0.01], [30.], [30.]])  # rho parameter of admm with L1-norm, TV and LLT prior
lam_array = np.array([[2], [2], [2]])  # relative weight of L1-norm, TV and LLT term
num_iters = 75

# Perform the denoising task using ADMM + 3 different priors
Poisson_denoiser(b, gt, lam_array, rho_array, num_iters, filename)
